@@ -46,16 +46,16 @@
 #include <limits.h>
 #include <math.h>
 
-#include "../../shared/util.h"
+#include <QtQuickTestUtils/private/qmlutils_p.h>
 
 class tst_qquickanimations : public QQmlDataTest
 {
     Q_OBJECT
 public:
-    tst_qquickanimations() {}
+    tst_qquickanimations() : QQmlDataTest(QT_QMLTEST_DATADIR) {}
 
 private slots:
-    void initTestCase()
+    void initTestCase() override
     {
         QQmlEngine engine;  // ensure types are registered
         QQmlDataTest::initTestCase();
@@ -117,9 +117,10 @@ private slots:
     void opacityAnimationFromZero();
     void alwaysRunToEndInSequentialAnimationBug();
     void cleanupWhenRenderThreadStops();
+    void infiniteLoopsWithoutFrom();
+    void targetsDeletedNotRemoved();
     void alwaysRunToEndSetFalseRestartBug();
 };
-
 #define QTIMED_COMPARE(lhs, rhs) do { \
     for (int ii = 0; ii < 5; ++ii) { \
         if (lhs == rhs)  \
@@ -219,7 +220,7 @@ void tst_qquickanimations::simpleColor()
     QVERIFY(animation.isPaused());
     animation.setCurrentTime(125);
     QCOMPARE(animation.currentTime(), 125);
-    QCOMPARE(rect.color(), QColor::fromRgbF(0.498039, 0, 0.498039, 1));
+    QCOMPARE(rect.color(), QColor::fromRgbF(0.498039f, 0, 0.498039f, 1));
 
     rect.setColor(QColor("green"));
     animation.setFrom(QColor("blue"));
@@ -230,7 +231,7 @@ void tst_qquickanimations::simpleColor()
     QCOMPARE(rect.color(), QColor("blue"));
     QVERIFY(animation.isRunning());
     animation.setCurrentTime(125);
-    QCOMPARE(rect.color(), QColor::fromRgbF(0.498039, 0, 0.498039, 1));
+    QCOMPARE(rect.color(), QColor::fromRgbF(0.498039f, 0, 0.498039f, 1));
 }
 
 void tst_qquickanimations::simpleRotation()
@@ -1237,7 +1238,7 @@ void tst_qquickanimations::easingProperties()
 
     {
         QQmlEngine engine;
-        QString componentStr = "import QtQuick 2.0\nPropertyAnimation { easing.type: \"Bezier\"; easing.bezierCurve: [0.5, 0.2, 0.13, 0.65, 1.0, 1.0] }";
+        QString componentStr = "import QtQuick 2.0\nPropertyAnimation { easing.type: \"BezierSpline\"; easing.bezierCurve: [0.5, 0.2, 0.13, 0.65, 1.0, 1.0] }";
         QQmlComponent animationComponent(&engine);
         animationComponent.setData(componentStr.toLatin1(), QUrl::fromLocalFile(""));
         QScopedPointer<QObject> obj(animationComponent.create());
@@ -1545,7 +1546,7 @@ void tst_qquickanimations::loopingBug()
     QVERIFY(anim != nullptr);
     QCOMPARE(anim->qtAnimation()->totalDuration(), 300);
     QCOMPARE(anim->isRunning(), true);
-    QTRY_COMPARE(static_cast<QAnimationGroupJob*>(anim->qtAnimation())->firstChild()->currentLoop(), 2);
+    QTRY_COMPARE(static_cast<QAnimationGroupJob*>(anim->qtAnimation())->children()->first()->currentLoop(), 2);
     QTRY_COMPARE(anim->isRunning(), false);
 
     QQuickRectangle *rect = obj->findChild<QQuickRectangle*>();
@@ -1900,7 +1901,7 @@ void tst_qquickanimations::opacityAnimationFromZero()
 
 void tst_qquickanimations::alwaysRunToEndInSequentialAnimationBug()
 {
-    QQuickView view(QUrl::fromLocalFile("data/alwaysRunToEndInSequentialAnimationBug.qml"));
+    QQuickView view(testFileUrl("alwaysRunToEndInSequentialAnimationBug.qml"));
     view.show();
 
     QVERIFY(QTest::qWaitForWindowExposed(&view));
@@ -2002,9 +2003,9 @@ void tst_qquickanimations::alwaysRunToEndInSequentialAnimationBug()
 
 void tst_qquickanimations::cleanupWhenRenderThreadStops()
 {
-    QQuickView view(QUrl::fromLocalFile("data/cleanupWhenRenderThreadStops.qml"));
+    QQuickView view(testFileUrl("cleanupWhenRenderThreadStops.qml"));
     view.show();
-    view.setPersistentOpenGLContext(false);
+    view.setPersistentGraphics(false);
     view.setPersistentSceneGraph(false);
     QVERIFY(QTest::qWaitForWindowExposed(&view));
     QTest::qWait(50);
@@ -2013,11 +2014,79 @@ void tst_qquickanimations::cleanupWhenRenderThreadStops()
     QVERIFY(QTest::qWaitForWindowExposed(&view));
 }
 
+void tst_qquickanimations::infiniteLoopsWithoutFrom()
+{
+    // This test checks QTBUG-84375
+    QQuickView view(testFileUrl("infiniteAnimationWithoutFrom.qml"));
+    view.show();
+
+    QVERIFY(QTest::qWaitForWindowExposed(&view));
+    QVERIFY(view.rootObject());
+
+    QObject *root = view.rootObject();
+    QQuickAbstractAnimation *animation = root->findChild<QQuickAbstractAnimation *>("anim");
+    QVERIFY(animation);
+    QQuickRectangle *rectangle = root->findChild<QQuickRectangle *>("rect");
+    QVERIFY(rectangle);
+
+    qreal prevRotation = rectangle->rotation();
+    int numsCrossedZero = 0;
+    connect(rectangle, &QQuickRectangle::rotationChanged, this, [&]() {
+        const auto rotation = rectangle->rotation();
+        // We take a large range of (180; 360] here, because the animation in
+        // the test runs for a short time, and so there can be huge gaps between
+        // rotation values.
+        const bool prevRotationOldLoop = prevRotation > 180.0 && prevRotation <= 360.0;
+        const bool currRotationNewLoop = rotation >= 0.0 && rotation <= 180.0;
+        if (prevRotationOldLoop && currRotationNewLoop)
+            numsCrossedZero++;
+        prevRotation = rotation;
+    });
+
+    // The logic in lamdba function above requires at least two positions in
+    // a rotation animation - one in [0; 180] range, and another in (180; 360]
+    // range
+    animation->start();
+    animation->pause();
+    QCOMPARE(numsCrossedZero, 0);
+    animation->setCurrentTime(40);
+    animation->setCurrentTime(90);
+    QCOMPARE(numsCrossedZero, 0);
+    animation->setCurrentTime(140);
+    animation->setCurrentTime(190);
+    QCOMPARE(numsCrossedZero, 1);
+    animation->setCurrentTime(240);
+    animation->setCurrentTime(290);
+    QCOMPARE(numsCrossedZero, 2);
+
+    animation->stop();
+}
+
+void tst_qquickanimations::targetsDeletedNotRemoved()
+{
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("targetsDeletedWithoutRemoval.qml"));
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY2(obj.get(), qPrintable(component.errorString()));
+    {
+        QQmlListReference ref(obj.get(), "targets");
+        QVERIFY(ref.isValid());
+        QCOMPARE(ref.size(), 1);
+        QTRY_COMPARE(ref.at(0), nullptr);
+    }
+    {
+        QQmlListReference ref(obj.get(), "animTargets");
+        QVERIFY(ref.isValid());
+        QCOMPARE(ref.size(), 1);
+        QCOMPARE(ref.at(0), nullptr);
+    }
+}
+
 //QTBUG-125224
 void tst_qquickanimations::alwaysRunToEndSetFalseRestartBug()
 {
     QQuickRectangle rect;
-    QQuickAnimationGroup sequential(nullptr);
+    QQuickSequentialAnimation sequential;
     QQuickPropertyAnimation beginAnim;
     QQuickPropertyAnimation endAnim;
 
